using System;
using System.Net;
using System.Net.Sockets;
using System.Text;
using GatewaySCO;
using Serilog;

namespace Servidor.TcpGateway;

public class TcpGateway
{
    private const int GatewayPort = 6690;
    private const string PosbcServer = "127.0.0.1";
    private const int PosbcPort = 6697;

    public async Task Start()
    {
        try
        {
            TcpListener gatewayListener = new TcpListener(IPAddress.Any, GatewayPort);
            gatewayListener.Start();
            Log.Information($"Gateway B#1 escuchando en el puerto {GatewayPort}...");

            while (true)
            {
                TcpClient checClient = await gatewayListener.AcceptTcpClientAsync();
                Log.Information("CHEC conectado.");
                _ = Task.Run(() => HandleChecClientAsync(checClient));
            }
        }
        catch (Exception ex)
        {
            Log.Error(ex, "Error en el Gateway");
        }
        finally
        {
            Log.CloseAndFlush();
        }
    }

    private async Task HandleChecClientAsync(TcpClient checClient)
    {
        using (checClient)
        using (NetworkStream checStream = checClient.GetStream())
        using (TcpClient posbcClient = new TcpClient())
        {
            try
            {
                await posbcClient.ConnectAsync(PosbcServer, PosbcPort);
                Log.Information("Conectado a POSBC.");
                using (NetworkStream posbcStream = posbcClient.GetStream())
                {
                    Task receiveFromPosbcTask = ReceiveFromPosbcAndSendToChecAsync(posbcStream, checStream);

                    while (true)
                    {
                        // Leer mensaje desde CHEC
                        byte[] lengthBuffer = new byte[4];
                        int bytesRead = await checStream.ReadAsync(lengthBuffer, 0, 4);
                        if (bytesRead == 0) continue; // Si no se lee nada, continuar esperando

                        int messageLength = Util.LongitudCodificada(lengthBuffer);
                        Log.Debug("Mensaje de CHEC longitud {longitud}", messageLength);
                        byte[] messageBuffer = new byte[messageLength];
                        bytesRead = await checStream.ReadAsync(messageBuffer, 0, messageLength);
                        if (bytesRead == 0) continue; // Si no se lee nada, continuar esperando

                        // Remitir mensaje a POSBC
                        byte[] bytesMsj = Util.ConcatenaArreglosBytes(lengthBuffer, messageBuffer);
                        // await posbcStream.WriteAsync(lengthBuffer, 0, 4);
                        // await posbcStream.WriteAsync(messageBuffer, 0, messageLength);
                        await posbcStream.WriteAsync(bytesMsj, 0, bytesMsj.Length);

                        Log.Information($"Mensaje remitido a POSBC: {Encoding.UTF8.GetString(messageBuffer)}");
                    }

                    await receiveFromPosbcTask;
                }
            }
            catch (Exception ex)
            {
                Log.Error(ex, "Error manejando la conexión CHEC");
            }
        }
    }

    private async Task ReceiveFromPosbcAndSendToChecAsync(NetworkStream posbcStream, NetworkStream checStream)
    {
        while (true)
        {
            try
            {
                // Leer longitud del mensaje desde POSBC
                byte[] lengthBuffer = new byte[4];
                int bytesRead = await posbcStream.ReadAsync(lengthBuffer, 0, 4);
                if (bytesRead == 0) continue; // Si no se lee nada, continuar esperando

                int messageLength = Util.LongitudCodificada(lengthBuffer);
                Log.Debug("POSBC bytes cabecera leidos {bytes}, longitud en cabecera {longitud}", bytesRead, messageLength);
                byte[] messageBuffer = new byte[messageLength];
                bytesRead = await posbcStream.ReadAsync(messageBuffer, 0, messageLength);
                Log.Debug("POSBC bytes cuerpo ledidos {bytes}", bytesRead);
                if (bytesRead == 0) continue; // Si no se lee nada, continuar esperando

                // Remitir mensaje a CHEC
                byte[] bytesMsj = Util.ConcatenaArreglosBytes(lengthBuffer, messageBuffer);
                // await checStream.WriteAsync(lengthBuffer, 0, 4);
                // await checStream.WriteAsync(messageBuffer, 0, messageLength);
                await checStream.WriteAsync(bytesMsj, 0, bytesMsj.Length);

                Log.Information($"Mensaje remitido a CHEC: {Encoding.UTF8.GetString(messageBuffer)}");
            }
            catch (Exception ex)
            {
                Log.Error(ex, "Error recibiendo desde POSBC o enviando a CHEC");
                break;
            }
        }
    }
}
